#ifndef _GRF4GENOBJECT_H_
#define _GRF4GENOBJECT_H_

PRISM *gdrGenDisc(/* int, float, float */);
PRISM *gdrGenCone(/* int, float, float, float */);
PRISM *gdrGenPrism(/* int, float*, float*, char*, float */);
SPHERE *gdrGenSphere(/* int, float */);
POLYGON *gdrGenPolygon(/* int, float* */);
POLYGON *gdrGenRegPolygon(/* int, float */);
LINESET *gdrGenLineSet(/* int, float* */);

double gdrSetDepth (/* Grf4Object*, float */);
void gdrDrawPrism (/* Grf4Object* */);
void gdrDrawSphere (/* Grf4Object* */);
void gdrDrawPolygon (/* Grf4Object* */);
void gdrDrawLineSet (/* Grf4Object* */);

int gdrOverPrism(/* float*, Grf4Object*, float, float */);
int gdrOverSphere(/* float*, Grf4Object*, float, float */);

#endif
